/*
 *	      PPP OS Layer Interface Module
 *
 *	    Written by Toshiharu OHNO (tony-o@iij.ad.jp)
 *
 *   Copyright (C) 1993, Internet Initiative Japan, Inc. All rights reserverd.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the Internet Initiative Japan, Inc.  The name of the
 * IIJ may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include "fsm.h"
#include <sys/param.h>
#include <sys/socket.h>
#include <net/route.h>
#if BSD >= 199206 || _BSDI_VERSION >= 199312
#include <sys/select.h>
#endif
#include <sys/ioctl.h>
#include <fcntl.h>
#include <net/if.h>
#ifdef __hpux
#include <net/if_ni.h>
#include <sys/uio.h>
#else
#include <net/if_tun.h>
#endif
#include <errno.h>
#include "ipcp.h"
#include "vars.h"
#include "os.h"

#ifdef __hpux
struct ifaliasreq {
  char    ifra_name[IFNAMSIZ];            /* if name, e.g. "en0" */
  struct  sockaddr ifra_addr;
  struct  sockaddr ifra_broadaddr;
  struct  sockaddr ifra_mask;
};
#endif

static struct ifaliasreq ifra;
static struct ifreq ifrq;
static struct in_addr oldmine, oldhis;
static int linkup;

#ifdef bsdi
extern char *inet_ntoa();
#endif
extern void HangupModem();

#ifdef __hpux
extern char IfName[IFNAMSIZ];
#endif
char *IfDevName;

#ifdef __hpux
static int
SetIfroute(add, dst, gateway)
int add;
struct in_addr dst, gateway;
{
  int s;
  struct rtentry    rt;

  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0) {
    perror("socket");
    return(-1);
  }

  bzero(&rt, sizeof(rt));
  rt.rt_dst.sa_family = rt.rt_gateway.sa_family = AF_INET;
  ((struct sockaddr_in *)&rt.rt_dst)->sin_addr.s_addr = dst.s_addr;
  ((struct sockaddr_in *)&rt.rt_gateway)->sin_addr.s_addr = gateway.s_addr;
  if (add) {
    rt.rt_flags = RTF_UP | RTF_HOST;
    if (ioctl(s, SIOCADDRT, &rt) < 0)
      perror("SIOCADDRT");
  }
  else {
    rt.rt_flags = RTF_HOST;
    if (ioctl(s, SIOCDELRT, &rt) < 0)
      perror("SIOCDELRT");
  }
  close(s);
  return(0);
}

static int
SetIfaddr()
{
  int s;

  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0) {
    perror("socket");
    return(-1);
  }

  bcopy(&ifra.ifra_addr, &ifrq.ifr_addr, sizeof(struct sockaddr));
  if (ioctl(s, SIOCSIFADDR, &ifrq) < 0)
    perror( "SIOCSIFADDR");

  /* Need to do this since HP-UX 10.10 'helpfully' creates bad routes */
  DeleteIfRoutes(1);

  bcopy(&ifra.ifra_broadaddr, &ifrq.ifr_dstaddr, sizeof(struct sockaddr));
  if (ioctl(s, SIOCSIFDSTADDR, &ifrq) < 0)
    perror( "SIOCSIFDSTADDR");

  bcopy(&ifra.ifra_mask, &ifrq.ifr_broadaddr, sizeof(struct sockaddr));
  if (ioctl(s, SIOCSIFNETMASK, &ifrq) < 0)
    perror( "SIOCSIFNETMASK");

  close(s);
  return(0);
}

static int
SetIpDevice(myaddr, hisaddr, netmask, updown)
struct in_addr myaddr, hisaddr, netmask;
int updown;
{
  struct sockaddr_in *sin;
  u_long mask, addr;

  if (updown == 0) {
    if (oldmine.s_addr == 0 && oldhis.s_addr == 0)
      return(0);
    bzero(&ifra.ifra_addr, sizeof(ifra.ifra_addr));
    bzero(&ifra.ifra_broadaddr, sizeof(ifra.ifra_addr));
    bzero(&ifra.ifra_mask, sizeof(ifra.ifra_addr));
#ifdef DEBUG
    logprintf("DIFADDR\n");
#endif
    SetIfroute(0, oldhis, oldmine);
    SetIfaddr();

    oldmine.s_addr = oldhis.s_addr = 0;
  } else {
    /*
     * If given addresses are alreay set, then ignore this request.
     */
    if (oldmine.s_addr == myaddr.s_addr && oldhis.s_addr == hisaddr.s_addr)
      return(0);
    /*
     * If different address has been set, then delete it first.
     */
    if (oldmine.s_addr || oldhis.s_addr) {
      SetIfroute(0, oldhis, oldmine);
    }
    /*
     *  Set interface address
     */
    sin = (struct sockaddr_in *)&(ifra.ifra_addr);
    sin->sin_family = AF_INET;
    sin->sin_addr = oldmine = myaddr;
    /*
     *  Set destination address
     */
    sin = (struct sockaddr_in *)&(ifra.ifra_broadaddr);
    sin->sin_family = AF_INET;
    sin->sin_addr = oldhis = hisaddr;
    /*
     */
    addr = ntohl(myaddr.s_addr);
    if (IN_CLASSA(addr))
      mask = IN_CLASSA_NET;
    else if (IN_CLASSB(addr))
      mask = IN_CLASSB_NET;
    else
      mask = IN_CLASSC_NET;
    /*
     *  if subnet mask is given, use it instead of class mask.
     */
    if (netmask.s_addr && (ntohl(netmask.s_addr) & mask) == mask)
      mask = ntohl(netmask.s_addr);
    sin = (struct sockaddr_in *)&(ifra.ifra_mask);
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = htonl(mask);

    SetIfaddr();
    SetIfroute(1, hisaddr, myaddr);
  }
  return(0);
}

#else /* __hpux */

static int
SetIpDevice(myaddr, hisaddr, netmask, updown)
struct in_addr myaddr, hisaddr, netmask;
int updown;
{
  struct sockaddr_in *sin;
  int s;
  int changeaddr = 0;
  u_long mask, addr;

  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0) {
    perror("socket");
    return(-1);
  }

  if (updown == 0) {
    if (oldmine.s_addr == 0 && oldhis.s_addr == 0) {
      return(0);
    }
    bzero(&ifra.ifra_addr, sizeof(ifra.ifra_addr));
    bzero(&ifra.ifra_broadaddr, sizeof(ifra.ifra_addr));
    bzero(&ifra.ifra_mask, sizeof(ifra.ifra_addr));
#ifdef DEBUG
    logprintf("DIFADDR\n");
#endif
    if (ioctl(s, SIOCDIFADDR, &ifra) < 0) {
      perror("SIOCDIFADDR");
      return(-1);
    }

    oldmine.s_addr = oldhis.s_addr = 0;
  } else {
    /*
     * If given addresses are alreay set, then ignore this request.
     */
    if (oldmine.s_addr == myaddr.s_addr && oldhis.s_addr == hisaddr.s_addr)
      return(0);
    /*
     * If different address has been set, then delete it first.
     */
    if (oldmine.s_addr || oldhis.s_addr) {
      changeaddr = 1;
    }
    /*
     *  Set interface address
     */
    sin = (struct sockaddr_in *)&(ifra.ifra_addr);
    sin->sin_family = AF_INET;
    sin->sin_addr = oldmine = myaddr;
    sin->sin_len = sizeof(*sin);
    /*
     *  Set destination address
     */
    sin = (struct sockaddr_in *)&(ifra.ifra_broadaddr);
    sin->sin_family = AF_INET;
    sin->sin_addr = oldhis = hisaddr;
    sin->sin_len = sizeof(*sin);
    /*
     */
    addr = ntohl(myaddr.s_addr);
    if (IN_CLASSA(addr))
      mask = IN_CLASSA_NET;
    else if (IN_CLASSB(addr))
      mask = IN_CLASSB_NET;
    else
      mask = IN_CLASSC_NET;
    /*
     *  if subnet mask is given, use it instead of class mask.
     */
    if (netmask.s_addr && (ntohl(netmask.s_addr) & mask) == mask)
      mask = ntohl(netmask.s_addr);

    sin = (struct sockaddr_in *)&(ifra.ifra_mask);
    sin->sin_family = AF_INET;
    sin->sin_addr.s_addr = htonl(mask);
    sin->sin_len = sizeof(*sin);

    if (changeaddr) {
      /*
       * Interface already exists. Just change the address.
       */
      bcopy(&ifra.ifra_addr, &ifrq.ifr_addr, sizeof(struct sockaddr));
      if (ioctl(s, SIOCSIFADDR, &ifra) < 0)
	perror("SIFADDR");;
      bcopy(&ifra.ifra_broadaddr, &ifrq.ifr_dstaddr, sizeof(struct sockaddr));
      if (ioctl(s, SIOCSIFDSTADDR, &ifrq) < 0)
	perror("SIFDSTADDR");;
#ifdef notdef
      bcopy(&ifra.ifra_mask, &ifrq.ifr_broadaddr, sizeof(struct sockaddr));
      if (ioctl(s, SIOCSIFBRDADDR, &ifrq) < 0)
	perror("SIFBRDADDR");
#endif
    } else if (ioctl(s, SIOCAIFADDR, &ifra) < 0) {
      perror("SIOCAIFADDR");
      return(-1);
    }
  }
  close(s);
  return(0);
}

#endif /* __hpux */

int
OsSetIpaddress(myaddr, hisaddr, netmask)
struct in_addr myaddr, hisaddr, netmask;
{
  return(SetIpDevice(myaddr, hisaddr, netmask, 1));
}

static struct in_addr peer_addr;
struct in_addr defaddr;

void
OsLinkup()
{
  char *s;

  if (linkup == 0) {
    if (setuid(0) < 0)
	logprintf("setuid failed\n");
    peer_addr = IpcpInfo.his_ipaddr;
    s = (char *)inet_ntoa(peer_addr);
    LogPrintf(LOG_PHASE, "OsLinkup: %s\n", s);

    if (SelectSystem(inet_ntoa(IpcpInfo.want_ipaddr), LINKFILE) < 0) {
      if (dstsystem) {
        if (SelectSystem(dstsystem, LINKFILE) < 0)
          SelectSystem("MYADDR", LINKFILE);
      } else
        SelectSystem("MYADDR", LINKFILE);
    }
    linkup = 1;
  }
  OsProxyarp(peer_addr, 1);
}

void
OsLinkdown()
{
  char *s;

  if (linkup) {
    s = (char *)inet_ntoa(peer_addr);
    LogPrintf(LOG_PHASE, "OsLinkdown: %s\n", s);
    if (!(mode & MODE_AUTO))
      DeleteIfRoutes(0);
    linkup = 0;
  }
  OsProxyarp(peer_addr, 0);
}

int
OsProxyarp(addr, install)
struct in_addr addr;
int install;
{
  static int installed = 0;
  int s;
  int i;
  int valhi, vallo;
  struct arpreq arpreq;

  if (!*VarProxyarp)
    return(0);

  s = socket(AF_INET, SOCK_DGRAM, 0);
  bzero(&arpreq, sizeof arpreq);
  arpreq.arp_pa.sa_family = AF_INET;
  ((struct sockaddr_in *)&arpreq.arp_pa)->sin_addr = addr;

  if (!install && installed) {
#ifdef DEBUG
    logprintf("ioctl SIOCDARP: %08x\n", addr);
#endif
    if (ioctl(s, SIOCDARP, (caddr_t)&arpreq) < 0) {
      perror("ioctl SIOCDARP");
      close(s);
      return(-1);
    }
    else
      installed = 0;
  } else if (install && !installed) {
    arpreq.arp_flags = ATF_PERM | ATF_PUBL;
    arpreq.arp_ha.sa_family = AF_UNSPEC;
    for (i = 0; i < 12; i++) {
      valhi = VarProxyarp[i] - '0' - (VarProxyarp[i] > '9') * ('a' - '9' - 1);
      i++;
      vallo = VarProxyarp[i] - '0' - (VarProxyarp[i] > '9') * ('a' - '9' - 1);
      arpreq.arp_ha.sa_data[i / 2] = (valhi << 4) + vallo;
    }
#ifdef DEBUG
    logprintf("ioctl SIOCSARP: %08x %02x:%02x:%02x:%02x:%02x:%02x\n", addr.s_addr, (unsigned int)arpreq.arp_ha.sa_data[0], (unsigned int)arpreq.arp_ha.sa_data[1], (unsigned int)arpreq.arp_ha.sa_data[2], (unsigned int)arpreq.arp_ha.sa_data[3], (unsigned int)arpreq.arp_ha.sa_data[4], (unsigned int)arpreq.arp_ha.sa_data[5]);
#endif
    if (ioctl(s, SIOCSARP, (caddr_t)&arpreq) < 0) {
      perror("ioctl SIOCSARP");
      close(s);
      return(-1);
    }
    else
      installed = 1;
  }
  close(s);
  return(0);
}

int
OsInterfaceDown(final)
int final;
{
  struct in_addr zeroaddr;
  int s;

  OsLinkdown();
  if (!final && (mode & MODE_AUTO))	/* We still want interface alive */
    return(0);
  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0) {
    perror("socket");
    return(-1);
  }
  ifrq.ifr_flags &= ~IFF_UP;
  if (ioctl(s, SIOCSIFFLAGS, &ifrq) < 0) {
    perror("SIOCSIFFLAGS");
    close(s);
    return(-1);
  }

  zeroaddr.s_addr = 0;
  SetIpDevice(zeroaddr, zeroaddr, zeroaddr, 0);

  close(s);
  return(0);
}

void
OsSetInterfaceParams(type, mtu, speed)
int type, mtu, speed;
{
#ifdef __hpux
  int tunflag;

  if (ioctl(tun_out, NIOCSMTYPE, &type) < 0)
    perror("NIOCSMTYPE");
  if (ioctl(tun_out, NIOCGFLAGS, &tunflag) < 0)
    perror("NIOCGFLAGS");
  else {
    tunflag &= ~IFF_UP;
    if (ioctl(tun_out, NIOCSFLAGS, &tunflag) < 0)
      perror("NIOCSFLAGS");
    if (ioctl(tun_out, NIOCSMTU, &mtu) < 0)
      perror("NIOCSMTU");
    tunflag |= IFF_UP;
    if (ioctl(tun_out, NIOCSFLAGS, &tunflag) < 0)
      perror("NIOCSFLAGS");
  }
  if (ioctl(tun_out, NIOCSMSPEED, &speed) < 0)
    perror("NIOCSMSPEED");
#else /* __hpux */
  struct tuninfo info;

  info.type = type;
  info.mtu = mtu;
  info.baudrate = speed;
  if (ioctl(tun_out, TUNSIFINFO, &info) < 0)
    perror("TUNSIFINFO");
#endif /* __hpux */
}

/*
 *  Open tunnel device and returns its descriptor
 */
int
OpenTunnel(ptun)
int *ptun;
{
  int s;
#ifdef __hpux
  int tunflag;
#else
  char *cp;
  char *suffix = "0123456789";
#endif
  char ifname[IFNAMSIZ];
  char devname[12];

#ifdef __hpux
  strcpy(devname, "/dev/ni");
  tun_out = open(devname, O_RDWR);
  if (tun_out < 0) {
    fprintf(stderr, "No tunnel device is available.\n");
    return(-1);
  }
  /*
   * Set up tunnel device
   */
  tunflag = AF_INET;
  if (ioctl(tun_out, NIOCBIND, &tunflag) < 0) {
    perror("NIOCBIND");
    return (-1);
  }
  tunflag = 500;
  if (ioctl(tun_out, NIOCSMTU, &tunflag) < 0) {
    perror("NIOCSMTU");
    return (-1);
  }
  if (ioctl(tun_out, NIOCGFLAGS, &tunflag) < 0) {
    perror("NIOCGFLAGS");
    return (-1);
  }
  tunflag |= IFF_POINTOPOINT | IFF_NOTRAILERS;
  if (ioctl(tun_out, NIOCSFLAGS, &tunflag) < 0) {
    perror("NIOCSFLAGS");
    return (-1);
  }
  if (ioctl(tun_out, NIOCGUNIT, &tunflag) < 0) {
    perror("NIOCGUNIT");
    return (-1);
  }
  *ptun = tunflag;

  devname[7] = (char)(*ptun + 0x30);
  devname[8] = '\0';

#else /* __hpux */

  strcpy(devname, "/dev/tun0");
  for (cp = suffix; *cp; cp++) {
    devname[8] = *cp;
    tun_out = open(devname, O_RDWR);
    if (tun_out >= 0)
      break;
  }
  *ptun = cp - suffix;
  if (*cp == '\0') {
    fprintf(stderr, "No tunnel device is available.\n");
    return(-1);
  }

#endif /* __hpux */

  /*
   * At first, name the interface.
   */
  strcpy(ifname, devname + 5);

  bzero((char *)&ifra, sizeof(ifra));
  bzero((char *)&ifrq, sizeof(ifrq));

  strncpy(ifrq.ifr_name, ifname, IFNAMSIZ);
  strncpy(ifra.ifra_name, ifname, IFNAMSIZ);

  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0) {
    perror("socket");
    return(-1);
  }

  /*
   *  Now, bring up the interface.
   */
  if (ioctl(s, SIOCGIFFLAGS, &ifrq) < 0) {
    perror("SIOCGIFFLAGS");
    close(s);
    return(-1);
  }

  ifrq.ifr_flags |= IFF_UP;
  if (ioctl(s, SIOCSIFFLAGS, &ifrq) < 0) {
    perror("SIOCSIFFLAGS");
    close(s);
    return(-1);
  }

  tun_in = tun_out;
#ifdef __hpux
  /*
   * Comment here said "HP-UX 9.01 tunnel drive returns bad information by
   * select system call".  Does that mean this is unnecessary?  I assume it
   * is related to the select() bug comment in main.c.  Anyone know if this
   * is fixed in HP-UX 10.10?
   */
   fcntl(tun_in, F_SETFL, O_NDELAY);
#endif
  IfDevName = devname + 5;
  if (GetIfIndex(IfDevName) < 0) {
    fprintf(stderr, "can't find ifindex.\n");
    close(s);
    return(-1);
  }
#ifdef __hpux
  strcpy(IfName, IfDevName);
#endif
  printf("Using interface: %s\r\n", IfDevName);
  LogPrintf(LOG_PHASE, "Using interface: %s\n", IfDevName);
  close(s);
  return(0);
}

#ifdef __hpux
int
HPtun_read(fd, buf, sz)
int fd;
void *buf;
size_t sz;
{
  u_char dbuf[sizeof(struct sockaddr_in)];
  int nb;
  struct iovec rvec[2];

  rvec[0].iov_base = (caddr_t)dbuf;
  rvec[0].iov_len = sizeof(struct sockaddr_in);
  rvec[1].iov_base = buf;
  rvec[1].iov_len = sz;

  nb = readv(fd, rvec, 2);
  return (nb > sizeof(struct sockaddr_in)) ?
        nb - sizeof(struct sockaddr_in) : nb;
}

int
HPtun_write(fd, buf, sz)
int fd;
void *buf;
size_t sz;
{
  u_char dbuf[sizeof(struct sockaddr_in)];
  int nw;
  struct iovec wvec[2];

  wvec[0].iov_base = (caddr_t)dbuf;
  wvec[0].iov_len = sizeof(struct sockaddr_in);
  wvec[1].iov_base = buf;
  wvec[1].iov_len = sz;
  bzero(dbuf, sizeof(struct sockaddr_in));
  ((struct sockaddr_in *)dbuf)->sin_family = AF_INET;

  nw = writev(fd, wvec, 2);
  return (nw > sizeof(struct sockaddr_in)) ?
        nw - sizeof(struct sockaddr_in) : nw;
}

#endif /* __hpux */

void
OsCloseLink(flag)
int flag;
{
  HangupModem(flag);
}

void
OsAddInOctets(cnt)
int cnt;
{
}

void
OsAddOutOctets(cnt)
int cnt;
{
}

